import pytest

from class_registry import ClassRegistry
from class_registry.cache import ClassRegistryInstanceCache
from test import Bulbasaur, Charmander, Pokemon, Squirtle


@pytest.fixture(name="registry")
def fixture_registry() -> ClassRegistry[Pokemon]:
    registry = ClassRegistry[Pokemon](attr_name="element")
    registry.register(Bulbasaur)
    registry.register(Charmander)
    registry.register(Squirtle)
    return registry


@pytest.fixture(name="cache")
def fixture_cache(
    registry: ClassRegistry[Pokemon],
) -> ClassRegistryInstanceCache[Pokemon]:
    return ClassRegistryInstanceCache[Pokemon](registry)


def test_get(
    cache: ClassRegistryInstanceCache[Pokemon],
    registry: ClassRegistry[Pokemon],
):
    """
    When an instance is returned from
    :py:meth:`ClassRegistryInstanceCache.get`, future invocations return
    the same instance.
    """
    poke_1 = cache["grass"]
    assert isinstance(poke_1, Bulbasaur)

    # Same key = exact same instance.
    poke_2 = cache["grass"]
    assert poke_2 is poke_1

    poke_3 = cache["water"]
    assert isinstance(poke_3, Squirtle)

    # If we pull a class directly from the wrapped registry, we get
    # a new instance.
    poke_4 = registry["water"]
    assert isinstance(poke_4, Squirtle)
    assert poke_3 is not poke_4


def test_template_args(registry: ClassRegistry[Pokemon]):
    """
    Extra params passed to the cache constructor are passed to the template
    function when creating new instances.
    """
    # Add an extra init param to the cache.
    cache = ClassRegistryInstanceCache(registry, name="Bruce")

    # The cache parameters are automatically applied to the class'
    # initializer.
    poke_1 = cache["fire"]
    assert isinstance(poke_1, Charmander)
    assert poke_1.name == "Bruce"

    poke_2 = cache["water"]
    assert isinstance(poke_2, Squirtle)
    assert poke_2.name == "Bruce"


def test_len(
    cache: ClassRegistryInstanceCache[Pokemon],
    registry: ClassRegistry[Pokemon],
):
    """
    Checking the length of a cache.
    """
    # The length only reflects the number of cached instances.
    assert len(cache) == 0

    # Calling ``__getitem__`` directly to sneak past the linter (:
    cache.__getitem__("water")
    cache.__getitem__("grass")

    assert len(cache) == 2



import sys
from importlib.metadata import DistributionFinder, PathDistribution
from os import path
from pathlib import Path


class DummyDistributionFinder(DistributionFinder):
    """
    Injects a dummy distribution into the meta path finder, so that we can
    pretend like it's been pip installed during unit tests (i.e., so that we
    can test ``EntryPointsClassRegistry``), without polluting the persistent
    virtualenv.
    """

    DUMMY_PACKAGE_DIR = "dummy_package.egg-info"

    @classmethod
    def install(cls):
        for finder in sys.meta_path:
            if isinstance(finder, cls):
                # If we've already installed an instance of the class, then
                # something is probably wrong with our tests.
                raise ValueError(f"{cls.__name__} is already installed")

        sys.meta_path.append(cls())

    @classmethod
    def uninstall(cls):
        for i, finder in enumerate(sys.meta_path):
            if isinstance(finder, cls):
                sys.meta_path.pop(i)
                return
        else:
            # If we didn't find an installed instance of the class, then
            # something is probably wrong with our tests.
            raise ValueError(f"{cls.__name__} was not installed")

    def find_distributions(self, context=...) -> list[PathDistribution]:
        return [
            PathDistribution(
                Path(path.join(path.dirname(__file__), self.DUMMY_PACKAGE_DIR))
            )
        ]



__all__ = ["AutoRegister"]

from abc import ABCMeta
from inspect import isabstract as is_abstract
from warnings import warn

from .base import BaseMutableRegistry


def AutoRegister(registry: BaseMutableRegistry, base_type: type = ABCMeta) -> type:
    """
    DEPRECATED: Use ``class_registry.base.AutoRegister`` instead (returns a base class
    instead of a metaclass).

    Creates a metaclass that automatically registers all non-abstract subclasses in the
    specified registry.

    Example::

       commands = ClassRegistry(attr_name='command_name')

       # Specify ``AutoRegister`` as the metaclass:
       class BaseCommand(metaclass=AutoRegister(commands)):
         @abstractmethod
         def print(self):
           raise NotImplementedError()

       class PrintCommand(BaseCommand):
         command_name = 'print'

         def print(self):
           ...

       print(list(commands.items())) # [('print', PrintCommand)]

    .. important::

       Python defines abstract as "having at least one unimplemented abstract method";
       adding :py:class:`abc.ABC` as a base class is not enough.

    :param registry:
        The registry that new classes will be added to.

        .. note::

           The registry's ``attr_name`` attribute must be set.

    :param base_type:
        The base type of the metaclass returned by this function.

        99.99% of the time, this should be :py:class:`abc.ABCMeta`.
    """
    warn(
        "class_registry.auto_register.AutoRegister is deprecated and will be removed in"
        "a future version of ClassRegistry.  Use class_registry.base.AutoRegister"
        "instead (returns a base class instead of a metaclass).  See"
        "https://github.com/todofixthis/class-registry/issues/14 for more information.",
        DeprecationWarning,
    )

    if not registry.attr_name:
        raise ValueError(f"Missing `attr_name` in {registry}.")

    class _metaclass(base_type):
        def __init__(self, what, bases=None, attrs=None):
            super().__init__(what, bases, attrs)

            if not is_abstract(self):
                registry.register(self)

    return _metaclass



__all__ = ["ClassRegistryInstanceCache"]

import typing
from collections import defaultdict

from . import ClassRegistry

T = typing.TypeVar("T")


class ClassRegistryInstanceCache(typing.Mapping[typing.Hashable, T]):
    """
    Wraps a ClassRegistry instance, caching instances as they are created.

    This allows you to create [multiple] registries that cache instances locally (so
    that they can be scoped and garbage-collected), whilst keeping the class registry
    separate.

    Note that the internal class registry is copied by reference, so any classes that
    are registered afterward are accessible to both the ``ClassRegistry`` and the
    ``ClassRegistryInstanceCache``.
    """

    def __init__(
        self,
        class_registry: ClassRegistry[T],
        *args,
        **kwargs,
    ) -> None:
        """
        :param class_registry:
            The wrapped ClassRegistry.

        :param args:
            Positional arguments passed to the class registry's template when creating
            new instances.

        :param kwargs:
            Keyword arguments passed to the class registry's template function when
            creating new instances.
        """
        super().__init__()

        self._registry: ClassRegistry[T] = class_registry
        self._cache: dict[typing.Hashable, T] = {}

        self._key_map: dict[typing.Hashable, list] = defaultdict(list)

        self._template_args = args
        self._template_kwargs = kwargs

    def __getitem__(self, key: typing.Hashable) -> T:
        """
        Returns the cached instance associated with the specified key.
        """
        instance_key = self.get_instance_key(key)

        if instance_key not in self._cache:
            class_key = self.get_class_key(key)

            # Map lookup keys to cache keys so that we can iterate over them in the
            # correct order.
            self._key_map[class_key].append(instance_key)

            self._cache[instance_key] = self._registry.get(
                class_key, *self._template_args, **self._template_kwargs
            )

        return self._cache[instance_key]

    def __iter__(self) -> typing.Generator[T, None, None]:
        """
        Returns a generator for iterating over cached instances, using the wrapped
        registry to determine sort order.

        If a key has not been accessed yet, it will not be included.
        """
        for lookup_key in self._registry.keys():
            for cache_key in self._key_map[lookup_key]:
                yield self._cache[cache_key]

    def __len__(self) -> int:
        """
        Returns the number of cached instances.

        Does not reflect any registered classes that haven't been accessed yet.
        """
        return len(self._cache)

    def warm_cache(self) -> None:
        """
        Warms up the cache, ensuring that an instance is created for every key currently
        in the registry.

        .. note::

           This method has no effect for any classes added to the wrapped
           :py:class:`ClassRegistry` after calling ``warm_cache``.
        """
        for key in self._registry.keys():
            self.__getitem__(key)

    def get_instance_key(self, key: typing.Hashable) -> typing.Hashable:
        """
        Generates a key that can be used to store/lookup values in the instance cache.

        :param key:
            Value provided to :py:meth:`__getitem__`.
        """
        return self.get_class_key(key)

    def get_class_key(self, key: typing.Hashable) -> typing.Hashable:
        """
        Generates a key that can be used to store/lookup values in the wrapped
        :py:class:`ClassRegistry` instance.

        This method is only invoked in the event of a cache miss.

        :param key:
            Value provided to :py:meth:`__getitem__`.
        """
        return self._registry.gen_lookup_key(key)



__all__ = ["AutoRegister", "BaseMutableRegistry", "BaseRegistry", "RegistryKeyError"]

import typing
from abc import ABC, abstractmethod as abstract_method
from collections.abc import Container
from inspect import isabstract as is_abstract, isclass as is_class
from warnings import warn


class RegistryKeyError(KeyError):
    """
    Used to differentiate a registry lookup from a standard KeyError.

    This is especially useful when a registry class expects to extract values from dicts
    to generate keys.
    """

    pass


T = typing.TypeVar("T")


class BaseRegistry(Container[T], ABC):
    """
    Base functionality for registries.
    """

    def __contains__(self, key: typing.Hashable) -> bool:
        """
        Returns whether the specified key is registered.
        """
        try:
            # Use :py:meth:`get_class` instead of :py:meth:`__getitem__`, to avoid
            # creating a new instance unnecessarily (i.e., prevent errors if the
            # corresponding class' constructor requires arguments).
            self.get_class(key)
        except RegistryKeyError:
            return False
        else:
            return True

    def __dir__(self) -> typing.Iterable[str]:
        """
        Attempts to return the list of registered keys.

        :raises: TypeError if a key cannot be cast as a string.
        """
        return list(map(str, self.keys()))

    def __getitem__(self, key: typing.Hashable) -> T:
        """
        Shortcut for calling :py:meth:`get` with empty args/kwargs.
        """
        return self.get(key)

    def __iter__(self) -> typing.Iterator[typing.Hashable]:
        """
        Iterates over registry keys.
        """
        return iter(self.keys())

    def __len__(self) -> int:
        """
        Returns the number of registered classes.
        """
        return sum(1 for _ in self.keys())

    def __missing__(self, key: typing.Hashable) -> typing.Type[T]:
        """
        Defines what to do when trying to access an unregistered key.

        Default behaviour is to throw a typed exception, but you could override this in
        a subclass, e.g., to return a default value.

        .. note::

           This method must return a class, not an instance.
        """
        raise RegistryKeyError(key)

    @abstract_method
    def get_class(self, key: typing.Hashable) -> typing.Type[T]:
        """
        Returns the class associated with the specified key.
        """
        raise NotImplementedError()

    def get(self, key: typing.Hashable, *args: typing.Any, **kwargs: typing.Any) -> T:
        """
        Creates a new instance of the class matching the specified key.

        :param key:
            The corresponding load key.

        :param args:
            Positional arguments passed to class initializer.
            Ignored if the class registry was initialized with a null template function.

        :param kwargs:
            Keyword arguments passed to class initializer.
            Ignored if the class registry was initialized with a null template function.

        References:
          - :py:meth:`__init__`
        """
        return self.create_instance(self.get_class(key), *args, **kwargs)

    @abstract_method
    def keys(self) -> typing.Iterable[typing.Hashable]:
        """
        Returns the collection of registered keys.
        """
        raise NotImplementedError()

    def classes(self) -> typing.Iterable[typing.Type[T]]:
        """
        Returns the collection of registered classes.
        """
        return iter(self.get_class(key) for key in self.keys())

    @staticmethod
    def gen_lookup_key(key: typing.Hashable) -> typing.Hashable:
        """
        Used by :py:meth:`get` to generate a lookup key.

        You may override this method in a subclass, for example if you need to support
        legacy aliases, etc.

        :param key: the key value provided to e.g., :py:meth:`__getitem__`
        :returns: the registry key, used to look up the corresponding class.
        """
        return key

    @staticmethod
    def create_instance(
        class_: typing.Type[T], *args: typing.Any, **kwargs: typing.Any
    ) -> T:
        """
        Prepares the return value for :py:meth:`get`.

        You may override this method in a subclass, if you want to customize the way new
        instances are created.

        :param class_:
            The requested class.

        :param args:
            Positional keywords passed to :py:meth:`get`.

        :param kwargs:
            Keyword arguments passed to :py:meth:`get`.
        """
        return class_(*args, **kwargs)


class BaseMutableRegistry(BaseRegistry[T], ABC):
    """
    Extends :py:class:`BaseRegistry` with methods that can be used to modify the
    registered classes.
    """

    def __init__(self, attr_name: typing.Optional[str] = None) -> None:
        """
        :param attr_name:
            If provided, :py:meth:`register` will automatically detect the key to use
            when registering new classes.
        """
        super().__init__()

        self.attr_name = attr_name

        # Map lookup keys to readable keys.
        # Only needed when :py:meth:`gen_lookup_key` is overridden, but I'm not good
        # enough at reflection black magic to figure out how to do that (:
        self._lookup_keys: dict[typing.Hashable, typing.Hashable] = {}

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.attr_name!r})"

    def keys(self) -> typing.Iterable[typing.Hashable]:
        """
        Returns the collection of registry keys, in the order that they were registered.
        """
        return iter(self._lookup_keys.keys())

    def items(self) -> typing.Iterable[tuple[typing.Hashable, typing.Type[T]]]:
        """
        DEPRECATED: use :py:meth:`keys` or :py:meth:`classes` instead.

        Returns the collection of registered key-class pairs, in the order that they
        were registered.
        """
        warn(
            f"{type(self).__name__}.items() is deprecated and will be removed in a "
            f"future version of ClassRegistry.  Use `zip({type(self).__name__}.keys(), "
            f"{type(self).__name__}.classes())` instead.",
            DeprecationWarning,
        )
        return iter(zip(self.keys(), self.classes()))

    def values(self) -> typing.Iterable[typing.Type[T]]:
        """
        DEPRECATED: use :py:meth:`classes` instead.

        Returns the collection of registered classes, in the order that they were
        registered.
        """
        warn(
            f"{type(self).__name__}.values() is deprecated and will be removed in a "
            f"future version of ClassRegistry.  Use {type(self).__name__}.classes()"
            f"instead.",
            DeprecationWarning,
        )
        return self.classes()

    if typing.TYPE_CHECKING:
        # :see: https://mypy.readthedocs.io/en/stable/generics.html#decorator-factories
        @typing.overload
        def register(self, key: typing.Type[T]) -> typing.Type[T]:
            """Decorator variant"""
            ...

        @typing.overload
        def register(
            self, key: typing.Hashable
        ) -> typing.Callable[[typing.Type[T]], typing.Type[T]]:
            """Decorator factory variant"""
            ...

    def register(
        self,
        key: typing.Union[typing.Hashable, typing.Type[T]],
    ) -> typing.Union[
        typing.Type[T],
        typing.Callable[[typing.Type[T]], typing.Type[T]],
    ]:
        """
        Decorator that registers a class with the registry.

        Example::

           registry = ClassRegistry(attr_name='widget_type')

           @registry.register
           class CustomWidget(BaseWidget):
             widget_type = 'custom'
             ...

           # Override the registry key:
           @registry.register('premium')
           class AdvancedWidget(BaseWidget):
             ...

        :param key:
            The registry key to use for the registered class.
            Optional if the registry's :py:attr:`attr_name` is set.
        """
        # ``@register`` usage:
        if is_class(key):
            if typing.TYPE_CHECKING:
                key = typing.cast(typing.Type[T], key)

            if self.attr_name:
                attr_key = getattr(key, self.attr_name)
                lookup_key = self.gen_lookup_key(attr_key)

                self._register(lookup_key, key)
                self._lookup_keys[attr_key] = lookup_key

                return key
            else:
                raise ValueError(
                    f"Attempting to register {key.__name__} to {type(self).__name__}"
                    f"via decorator, but `{type(self).__name__}.attr_key` is not set."
                )
        else:
            # :see: https://github.com/python/mypy/issues/16640
            if typing.TYPE_CHECKING:
                key = typing.cast(typing.Hashable, key)

            # ``@register('some_attr')`` usage:
            def _decorator(cls: typing.Type[T]) -> typing.Type[T]:
                lookup_key_ = self.gen_lookup_key(key)

                self._register(lookup_key_, cls)
                self._lookup_keys[key] = lookup_key_

                return cls

            return _decorator

    def unregister(self, key: typing.Hashable) -> typing.Type[T]:
        """
        Unregisters the class with the specified key.

        :param key:
            The registry key to remove (not the registered class!).

        :return:
            The class that was unregistered.

        :raise:
            - :py:class:`KeyError` if the key is not registered.
        """
        result = self._unregister(self.gen_lookup_key(key))
        del self._lookup_keys[key]

        return result

    @abstract_method
    def _register(self, key: typing.Hashable, class_: typing.Type[T]) -> None:
        """
        Registers a class with the registry.

        :param key: Return value from :py:meth:`gen_lookup_key`.
        """
        raise NotImplementedError()

    @abstract_method
    def _unregister(self, key: typing.Hashable) -> typing.Type[T]:
        """
        Unregisters the class at the specified key.

        :param key: Return value from :py:meth:`gen_lookup_key`.
        """
        raise NotImplementedError()


def AutoRegister(registry: BaseMutableRegistry) -> type:
    """
    Creates a base class that automatically registers all non-abstract subclasses in the
    specified registry.

    Example::

       commands = ClassRegistry(attr_name='command_name')

       class BaseCommand(AutoRegister(commands), ABC):
         @abstractmethod
         def print(self):
           raise NotImplementedError()

       class PrintCommand(BaseCommand):
         command_name = 'print'

         def print(self):
           ...

       print(list(commands.items())) # [('print', PrintCommand)]

    .. important::

       Python defines abstract as "having at least one unimplemented abstract method";
       adding :py:class:`abc.ABC` as a base class is not enough.

    :param registry:
        The registry that new classes will be added to.

        .. note::

           The registry's ``attr_name`` attribute must be set.
    """
    if not registry.attr_name:
        raise ValueError(f"Missing `attr_name` in {registry}.")

    class _Base:
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)

            if not is_abstract(cls):
                registry.register(cls)

    return _Base



import typing


class Pokemon:
    """
    A basic class with some attributes that we can use to test out class
    registries.
    """

    element: str

    def __init__(self, name: typing.Optional[str] = None):
        super().__init__()

        self.name: typing.Optional[str] = name


# Define some classes that we can register.
class Charmander(Pokemon):
    element = "fire"


class Charmeleon(Pokemon):
    element = "fire"


class Squirtle(Pokemon):
    element = "water"


class Wartortle(Pokemon):
    element = "water"


class Bulbasaur(Pokemon):
    element = "grass"


class Ivysaur(Pokemon):
    element = "grass"


class Mew(Pokemon):
    element = "psychic"


class PokemonFactory:
    """
    A factory that can produce new pokémon on demand.  Used to test how registries
    behave when a method/function is registered instead of a class.
    """

    @classmethod
    def create_psychic_pokemon(cls, name: typing.Optional[str] = None):
        return Mew(name)



"""
Verifies registry behaviour when :py:func:`class_registry.ClassRegistry.gen_lookup_key`
is modified.
"""
import typing

import pytest

from class_registry import ClassRegistry
from test import Charmander, Pokemon, Squirtle


@pytest.fixture(name="customised_registry")
def fixture_customised_registry() -> ClassRegistry[Pokemon]:
    class CustomisedLookupRegistry(ClassRegistry[Pokemon]):
        @staticmethod
        def gen_lookup_key(key: typing.Hashable) -> typing.Hashable:
            """
            Simple override of `gen_lookup_key`, to ensure the registry
            behaves as expected when the lookup key is different.
            """
            if isinstance(key, str):
                return "".join(reversed(key))
            return key

    registry = CustomisedLookupRegistry()
    registry.register("fire")(Charmander)
    registry.register("water")(Squirtle)
    return registry


def test_contains(customised_registry: ClassRegistry[Pokemon]):
    assert "fire" in customised_registry
    assert "erif" not in customised_registry


def test_getitem(customised_registry: ClassRegistry[Pokemon]):
    assert isinstance(customised_registry["fire"], Charmander)


def test_iter(customised_registry: ClassRegistry[Pokemon]):
    generator = iter(customised_registry)

    assert next(generator) == "fire"
    assert next(generator) == "water"

    with pytest.raises(StopIteration):
        next(generator)


def test_len(customised_registry: ClassRegistry[Pokemon]):
    assert len(customised_registry) == 2


def test_get_class(customised_registry: ClassRegistry[Pokemon]):
    assert customised_registry.get_class("fire") is Charmander


def test_get(customised_registry: ClassRegistry[Pokemon]):
    assert isinstance(customised_registry.get("fire"), Charmander)


def test_unregister(customised_registry: ClassRegistry[Pokemon]):
    customised_registry.unregister("fire")

    assert "fire" not in customised_registry
    assert "erif" not in customised_registry


def test_use_case_aliases():
    """
    A common use case for overriding `gen_lookup_key` is to specify some
    aliases (e.g., for backwards-compatibility when refactoring an existing
    registry).
    """

    class TestRegistry(ClassRegistry[Pokemon]):
        @staticmethod
        def gen_lookup_key(key: typing.Hashable) -> typing.Hashable:
            """
            Simulate a scenario where we renamed the key for a class in the
            registry, but we want to preserve backwards-compatibility with
            existing code that hasn't been updated yet.
            """
            if key == "bird":
                return "flying"

            return key

    registry = TestRegistry()

    @registry.register("flying")
    class MissingNo(Pokemon):
        pass

    assert isinstance(registry["bird"], MissingNo)
    assert isinstance(registry["flying"], MissingNo)

    assert "bird" in registry
    assert "flying" in registry



__all__ = ["ClassRegistry", "SortedClassRegistry"]

import typing
from functools import cmp_to_key

from .base import BaseMutableRegistry, RegistryKeyError

T = typing.TypeVar("T")


class ClassRegistry(BaseMutableRegistry[T]):
    """
    Maintains a registry of classes and provides a generic factory for instantiating
    them.
    """

    def __init__(
        self,
        attr_name: typing.Optional[str] = None,
        unique: bool = False,
    ) -> None:
        """
        :param attr_name:
            If provided, :py:meth:`register` will automatically detect the key to use
            when registering new classes.

        :param unique:
            Determines what happens when two classes are registered with the same key:

            - ``True``: A :py:class:`KeyError` will be raised.
            - ``False``: The second class will replace the first one.
        """
        super().__init__(attr_name)

        self.unique = unique

        self._registry: dict[typing.Hashable, typing.Type[T]] = {}

    def __len__(self) -> int:
        return len(self._registry)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(attr_name={self.attr_name!r}, unique={self.unique!r})"

    def get_class(self, key: typing.Hashable) -> typing.Type[T]:
        """
        Returns the class associated with the specified key.
        """
        lookup_key = self.gen_lookup_key(key)

        try:
            return self._registry[lookup_key]
        except KeyError:
            return self.__missing__(lookup_key)

    def _register(self, key: typing.Hashable, class_: typing.Type[T]) -> None:
        """
        Registers a class with the registry.

        :param key: Has already been processed by :py:meth:`gen_lookup_key`.
        """
        if key in ["", None]:
            raise ValueError(
                f"Attempting to register class {class_.__name__} "
                "with empty registry key {key!r}."
            )

        if self.unique and (key in self._registry):
            raise RegistryKeyError(
                f"{class_.__name__} with key {key!r} is already registered.",
            )

        self._registry[key] = class_

    def _unregister(self, key: typing.Hashable) -> typing.Type[T]:
        """
        Unregisters the class at the specified key.

        :param key: Has already been processed by :py:meth:`gen_lookup_key`.
        """
        return (
            self._registry.pop(key) if key in self._registry else self.__missing__(key)
        )


class SortedClassRegistry(ClassRegistry[T]):
    """
    A ClassRegistry that uses a function to determine sort order when iterating.
    """

    def __init__(
        self,
        sort_key: typing.Any,
        attr_name: typing.Optional[str] = None,
        unique: bool = False,
        reverse: bool = False,
    ) -> None:
        """
        :param sort_key:
            Attribute name or callable, used to determine the sort value.

            If callable, must accept two tuples of (key, class, lookup_key).

            You can also use :py:func:`functools.cmp_to_key`.

        :param attr_name:
            If provided, :py:meth:`register` will automatically detect the key to use
            when registering new classes.

        :param unique:
            Determines what happens when two classes are registered with the same key:

            - ``True``: The second class will replace the first one.
            - ``False``: A ``ValueError`` will be raised.

        :param reverse:
            Whether to reverse the sort ordering.
        """
        super().__init__(attr_name, unique)

        self._sort_key = (
            sort_key if callable(sort_key) else self.create_sorter(sort_key)
        )

        self.reverse = reverse

    def keys(self) -> typing.Iterable[typing.Hashable]:
        """
        Returns the collection of registry keys, in the order that they were registered.
        """
        return iter(
            key
            for key, _, _ in sorted(
                (
                    # Provide both human-readable and lookup keys to the sorter.
                    (key, self.get_class(key), self.gen_lookup_key(key))
                    for key in super().keys()
                ),
                key=self._sort_key,
                reverse=self.reverse,
            )
        )

    @staticmethod
    def create_sorter(sort_key: str):
        """
        Given a sort key, creates a function that can be used to sort items when
        iterating over the registry.
        """

        def sorter(a, b):
            a_attr = getattr(a[1], sort_key)
            b_attr = getattr(b[1], sort_key)

            return (a_attr > b_attr) - (a_attr < b_attr)

        return cmp_to_key(sorter)



import pytest

from class_registry import ClassRegistry, RegistryKeyError
from test import Bulbasaur, Charmander, Charmeleon, Pokemon, Squirtle, Wartortle


def test_register_manual_keys():
    """
    Registers a few classes with manually-assigned identifiers and verifies that the
    factory returns them correctly.
    """
    registry = ClassRegistry[Pokemon]()

    @registry.register("fire")
    class Charizard(Pokemon):
        pass

    class Blastoise(Pokemon):
        pass

    # Alternate syntax (rarely used; mostly just here to give mypy more to work with):
    registry.register("water")(Blastoise)

    # By default, you have to specify a registry key when registering new classes.
    # We'll see how to assign registry keys automatically in the next test.
    with pytest.raises(ValueError):
        # noinspection PyUnusedLocal
        @registry.register
        class Venusaur(Pokemon):
            pass

    assert registry.get_class("fire") is Charizard
    assert isinstance(registry["fire"], Charizard)

    assert registry.get_class("water") is Blastoise
    assert isinstance(registry["water"], Blastoise)


def test_register_detect_keys():
    """
    If an attribute name is passed to ClassRegistry's constructor, it will automatically
    check for this attribute when registering classes.
    """
    registry = ClassRegistry[Pokemon](attr_name="element")

    @registry.register
    class Charizard(Pokemon):
        element = "fire"

    class Blastoise(Pokemon):
        element = "water"

    # Alternate syntax (rarely used; mostly just here to give mypy more to work with):
    registry.register(Blastoise)

    # You can still override the registry key if you want.
    @registry.register("poison")
    class Venusaur(Pokemon):
        element = "grass"

    assert isinstance(registry["fire"], Charizard)
    assert isinstance(registry["water"], Blastoise)
    assert isinstance(registry["poison"], Venusaur)

    # We overrode the registry key for this class.
    with pytest.raises(RegistryKeyError):
        # noinspection PyStatementEffect
        registry["grass"]


def test_register_error_empty_key():
    """
    Attempting to register a class with an empty key.
    """
    registry = ClassRegistry[Pokemon]("element")

    with pytest.raises(ValueError):
        # noinspection PyUnusedLocal
        @registry.register("")
        class Rapidash(Pokemon):
            element = "fire"

    with pytest.raises(ValueError):
        # noinspection PyUnusedLocal
        @registry.register
        class Mew(Pokemon):
            element = None

    with pytest.raises(ValueError):
        # noinspection PyUnusedLocal
        @registry.register
        class Mewtwo(Pokemon):
            element = ""


def test_unique_keys():
    """
    Specifying ``unique=True`` when creating the registry requires all keys to be
    unique.
    """
    registry = ClassRegistry[Pokemon](attr_name="element", unique=True)

    # We can register any class like normal...
    registry.register(Charmander)

    # ... but if we try to register a second class with the same key, we
    # get an error.
    with pytest.raises(RegistryKeyError):
        registry.register(Charmeleon)


def test_unregister():
    """
    Removing a class from the registry.

    .. note::
       This is not used that often outside unit tests (e.g., to remove artefacts when a
       test has to add a class to a global registry).
    """
    registry = ClassRegistry[Pokemon](attr_name="element")
    registry.register(Charmander)
    registry.register(Squirtle)

    assert registry.unregister("fire") is Charmander

    with pytest.raises(RegistryKeyError):
        registry.get("fire")

    # Note that you must unregister the KEY, not the CLASS.
    with pytest.raises(KeyError):
        registry.unregister(Squirtle)

    # If you try to unregister a key that isn't registered, you'll
    # get an error.
    with pytest.raises(KeyError):
        registry.unregister("fire")


def test_constructor_params():
    """
    Params can be passed to the registered class' constructor.
    """
    registry = ClassRegistry[Pokemon](attr_name="element")
    registry.register(Bulbasaur)

    # Goofus uses positional arguments, which are magical and make his code more
    # difficult to read.
    goofus = registry.get("grass", "goofus")

    # Gallant uses keyword arguments, producing self-documenting code and being
    # courteous to his fellow developers.
    # He still names his pokémon after himself, though. Narcissist.
    gallant = registry.get("grass", name="gallant")

    assert isinstance(goofus, Bulbasaur)
    assert goofus.name == "goofus"

    assert isinstance(gallant, Bulbasaur)
    assert gallant.name == "gallant"


def test_new_instance_every_time():
    """
    Every time a registered class is invoked, a new instance is returned.
    """
    registry = ClassRegistry[Pokemon](attr_name="element")
    registry.register(Wartortle)

    assert registry["water"] is not registry["water"]


def test_register_function():
    """
    Functions can be registered as well (so long as they walk and quack like a class).
    """
    registry = ClassRegistry[Pokemon]()

    @registry.register("fire")
    def pokemon_factory(name=None):
        return Charmeleon(name=name)

    # Alternate syntax (rarely used; mostly just here to give mypy more to work with):
    # PyCharm doesn't like it, but mypy thinks it's fine :shrug:
    # noinspection PyTypeChecker
    registry.register("water")(pokemon_factory)

    poke1 = registry.get("fire", name="trogdor")
    assert isinstance(poke1, Charmeleon)
    assert poke1.name == "trogdor"

    poke2 = registry.get("water", name="leeroy")
    assert isinstance(poke2, Charmeleon)
    assert poke2.name == "leeroy"


def test_regression_contains_when_class_init_requires_arguments():
    """
    Special case when checking if a class is registered, and that class' initialiser
    requires arguments.
    """
    registry = ClassRegistry[Pokemon](attr_name="element")

    @registry.register
    class Butterfree(Pokemon):
        element = "bug"

        def __init__(self, name):
            super(Butterfree, self).__init__(name)

    # This line would raise a TypeError in a previous version of ClassRegistry.
    assert "bug" in registry



__all__ = ["EntryPointClassRegistry"]

import typing
from importlib.metadata import entry_points

from .base import BaseRegistry


T = typing.TypeVar("T")


class EntryPointClassRegistry(BaseRegistry[T]):
    """
    A class registry that loads classes using setuptools entry points.
    """

    def __init__(
        self,
        group: str,
        attr_name: typing.Optional[str] = None,
    ) -> None:
        """
        :param group:
            The name of the entry point group that will be used to load new classes.

        :param attr_name:
            If set, the registry will "brand" each class with its corresponding registry
            key.  This makes it easier to perform reverse lookups later.

            Note: if a class already defines this attribute, the registry will overwrite
            it!
        """
        super().__init__()

        self.attr_name = attr_name
        self.group = group

        self._cache: typing.Optional[dict[typing.Hashable, typing.Type[T]]] = None
        """
        Caches registered classes locally, so that we don't have to keep iterating over
        entry points.
        """

        # If :py:attr:`attr_name` is set, warm the cache immediately, to apply branding.
        if self.attr_name:
            self._get_cache()

    def __len__(self) -> int:
        return len(self._get_cache())

    def __repr__(self) -> str:
        return f"{type(self).__name__}(group={self.group!r})"

    def get(self, key: typing.Hashable, *args, **kwargs) -> T:
        instance = super().get(key, *args, **kwargs)

        if self.attr_name:
            # Apply branding to the instance explicitly.
            # This is particularly important if the corresponding entry point references
            # a function or method.
            setattr(instance, self.attr_name, key)

        return instance

    def get_class(self, key: typing.Hashable) -> typing.Type[T]:
        try:
            return self._get_cache()[key]
        except KeyError:
            return self.__missing__(key)

    def keys(self) -> typing.Iterable[typing.Hashable]:
        return iter(self._get_cache().keys())

    def refresh(self):
        """
        Purges the local cache.  The next access attempt will reload all entry points.

        This is useful if you load a distribution at runtime...such as during unit tests
        for ``phx-class-registry``.  Otherwise, it probably serves no useful purpose (:
        """
        self._cache = None

    def _get_cache(self) -> dict[typing.Hashable, typing.Type[T]]:
        """
        Populates the cache (if necessary) and returns it.
        """
        if self._cache is None:
            self._cache = {}
            for e in entry_points(group=self.group):
                cls = e.load()

                # Try to apply branding, but only for compatible types (i.e., functions
                # and methods can't be branded this way).
                if self.attr_name and isinstance(cls, type):
                    setattr(cls, self.attr_name, e.name)

                self._cache[e.name] = cls

        return self._cache



from importlib.metadata import entry_points

import pytest

from class_registry import RegistryKeyError
from class_registry.entry_points import EntryPointClassRegistry
from test import Bulbasaur, Charmander, Mew, Pokemon, PokemonFactory, Squirtle
from test.helper import DummyDistributionFinder


@pytest.fixture(name="distro", autouse=True)
def fixture_distro():
    # Inject a distribution that defines some entry points.
    DummyDistributionFinder.install()
    yield
    DummyDistributionFinder.uninstall()


def test_happy_path():
    """
    Loading classes automatically via entry points.

    See ``dummy_package.egg-info/entry_points.txt`` for more info.
    """
    registry = EntryPointClassRegistry[Pokemon]("pokemon")

    fire = registry["fire"]
    assert isinstance(fire, Charmander)
    assert fire.name is None

    grass = registry.get("grass")
    assert isinstance(grass, Bulbasaur)
    assert grass.name is None

    water = registry.get("water", name="archibald")
    assert isinstance(water, Squirtle)
    assert water.name == "archibald"

    # The 'psychic' entry point actually references a function, but it
    # works exactly the same as a class.
    psychic = registry.get("psychic", "snuggles")
    assert isinstance(psychic, Mew)
    assert psychic.name == "snuggles"


def test_branding():
    """
    Configuring the registry to "brand" each class/instance with its
    corresponding key.
    """
    registry = EntryPointClassRegistry[Pokemon]("pokemon", attr_name="poke_type")
    try:
        # Branding is applied immediately to each registered class.
        assert getattr(Charmander, "poke_type") == "fire"
        assert getattr(Squirtle, "poke_type") == "water"

        # Instances, too!
        assert getattr(registry["fire"], "poke_type") == "fire"
        assert getattr(registry.get("water", "phil"), "poke_type") == "water"

        # Registered functions and methods can't be branded this way,
        # though...
        assert not hasattr(PokemonFactory.create_psychic_pokemon, "poke_type")

        # ... but we can brand the resulting instances.
        assert getattr(registry["psychic"], "poke_type") == "psychic"
        assert getattr(registry.get("psychic"), "poke_type") == "psychic"
    finally:
        # Clean up after ourselves.
        for cls in registry.classes():
            if isinstance(cls, type):
                try:
                    delattr(cls, "poke_type")
                except AttributeError:
                    pass


def test_len():
    """
    Getting the length of an entry point class registry.
    """
    # Just in case some other package defines pokémon entry
    # points (:
    expected = len(list(entry_points(group="pokemon")))

    # Quick sanity check, to make sure our test pokémon are
    # registered correctly.
    assert expected >= 4

    registry = EntryPointClassRegistry[Pokemon]("pokemon")
    assert len(registry) == expected


def test_error_wrong_group():
    """
    The registry can't find entry points associated with the wrong group.
    """
    registry = EntryPointClassRegistry[Pokemon]("fhqwhgads")

    with pytest.raises(RegistryKeyError):
        registry.get("fire")



"""
Unit tests for the deprecated :py:class:`class_registry.auto_register.AutoRegister`
function.

This function is deprecated; use :py:class:`class_registry.base.AutoRegister` instead.
:see: https://github.com/todofixthis/class-registry/issues/14
"""
from abc import ABC, abstractmethod as abstract_method

import pytest

from class_registry import ClassRegistry

# noinspection PyDeprecation
from class_registry.auto_register import AutoRegister


def test_auto_register():
    """
    Using :py:func:`AutoRegister` to, well, auto-register classes.
    """
    registry = ClassRegistry(attr_name="element")

    # :py:func:`AutoRegister` is deprecated.
    # :see: https://github.com/todofixthis/class-registry/issues/14
    with pytest.deprecated_call():
        # Note that we declare :py:func:`AutoRegister` as the metaclass
        # for our base class.
        # noinspection PyDeprecation
        class BasePokemon(metaclass=AutoRegister(registry)):
            """
            Abstract base class; will not get registered.
            """

            @abstract_method
            def get_abilities(self):
                raise NotImplementedError()

    class Sandslash(BasePokemon):
        """
        Non-abstract subclass; will get registered automatically.
        """

        element = "ground"

        def get_abilities(self):
            return ["sand veil"]

    class BaseEvolvingPokemon(BasePokemon, ABC):
        """
        Abstract subclass; will not get registered.
        """

        @abstract_method
        def evolve(self):
            raise NotImplementedError()

    class Ekans(BaseEvolvingPokemon):
        """
        Non-abstract subclass; will get registered automatically.
        """

        element = "poison"

        def get_abilities(self):
            return ["intimidate", "shed skin"]

        def evolve(self):
            return "Congratulations! Your EKANS evolved into ARBOK!"

    # Note that only non-abstract classes got registered.
    assert list(registry.classes()) == [Sandslash, Ekans]


def test_abstract_strict_definition():
    """
    If a class has no unimplemented abstract methods, it gets registered.
    """
    registry = ClassRegistry(attr_name="element")

    # :py:func:`AutoRegister` is deprecated.
    # :see: https://github.com/todofixthis/class-registry/issues/14
    with pytest.deprecated_call():
        # noinspection PyDeprecation
        class FightingPokemon(metaclass=AutoRegister(registry)):
            element = "fighting"

    # :py:class:`FightingPokemon` does not define any abstract methods, so it is not
    # considered to be abstract!
    assert list(registry.classes()) == [FightingPokemon]


def test_error_attr_name_missing():
    """
    The registry doesn't have an ``attr_name``.
    """
    registry = ClassRegistry()

    with pytest.raises(ValueError):
        # :py:func:`AutoRegister` is deprecated.
        # :see: https://github.com/todofixthis/class-registry/issues/14
        with pytest.deprecated_call():
            # noinspection PyDeprecation
            AutoRegister(registry)



import pytest

from class_registry import ClassRegistry, RegistryKeyError
from class_registry.patcher import RegistryPatcher
from test import Bulbasaur, Charmander, Charmeleon, Ivysaur, Pokemon, Squirtle


@pytest.fixture(name="registry")
def fixture_registry() -> ClassRegistry[Pokemon]:
    return ClassRegistry(attr_name="element", unique=True)


def test_patch_detect_keys(registry: ClassRegistry[Pokemon]):
    """
    Patching a registry in a context, with registry keys detected
    automatically.
    """
    registry.register(Charmander)
    registry.register(Squirtle)

    with RegistryPatcher(registry, Charmeleon, Bulbasaur):
        assert isinstance(registry["fire"], Charmeleon)
        assert isinstance(registry["water"], Squirtle)

        # Nesting contexts?  You betcha!
        with RegistryPatcher(registry, Ivysaur):
            assert isinstance(registry["grass"], Ivysaur)

        assert isinstance(registry["grass"], Bulbasaur)

    # Save file corrupted.  Restoring previous save...
    assert isinstance(registry["fire"], Charmander)
    assert isinstance(registry["water"], Squirtle)

    with pytest.raises(RegistryKeyError):
        registry.get("grass")


def test_patch_manual_keys(registry: ClassRegistry[Pokemon]):
    """
    Patching a registry in a context, specifying registry keys manually.
    """
    registry.register("sparky")(Charmander)
    registry.register("chad")(Squirtle)

    with RegistryPatcher(registry, sparky=Charmeleon, rex=Bulbasaur):
        assert isinstance(registry["sparky"], Charmeleon)
        assert isinstance(registry["chad"], Squirtle)

        # Don't worry Chad; your day will come!
        with RegistryPatcher(registry, rex=Ivysaur):
            assert isinstance(registry["rex"], Ivysaur)

        assert isinstance(registry["rex"], Bulbasaur)

    # Save file corrupted.  Restoring previous save...
    assert isinstance(registry["sparky"], Charmander)
    assert isinstance(registry["chad"], Squirtle)

    with pytest.raises(RegistryKeyError):
        registry.get("jodie")



import typing
from functools import cmp_to_key

from class_registry.registry import SortedClassRegistry
from test import Bulbasaur, Charmander, Pokemon, Squirtle


def test_sort_key():
    """
    When iterating over a SortedClassRegistry, classes are returned in
    sorted order rather than inclusion order.
    """
    registry = SortedClassRegistry(
        attr_name="element",
        sort_key="weight",
    )

    @registry.register
    class Geodude(Pokemon):
        element = "rock"
        weight = 100

    @registry.register
    class Machop(Pokemon):
        element = "fighting"
        weight = 75

    @registry.register
    class Bellsprout(Pokemon):
        element = "grass"
        weight = 15

    # The registry iterates over registered classes in ascending order by
    # ``weight``.
    assert list(registry.classes()) == [Bellsprout, Machop, Geodude]


def test_sort_key_reverse():
    """
    Reversing the order of a sort key.
    """
    registry = SortedClassRegistry(
        attr_name="element",
        sort_key="weight",
        reverse=True,
    )

    @registry.register
    class Geodude(Pokemon):
        element = "rock"
        weight = 100

    @registry.register
    class Machop(Pokemon):
        element = "fighting"
        weight = 75

    @registry.register
    class Bellsprout(Pokemon):
        element = "grass"
        weight = 15

    # The registry iterates over registered classes in descending order by ``weight``.
    assert list(registry.classes()) == [Geodude, Machop, Bellsprout]


def test_cmp_to_key():
    """
    If you want to use a ``cmp`` function to define the ordering,
    you must use the :py:func:`cmp_to_key` function.
    """

    def compare_pokemon(a, b):
        """
        Sort in descending order by popularity.

        :param a: Tuple of (key, class, lookup_key)
        :param b: Tuple of (key, class, lookup_key)
        """
        return (a[1].popularity < b[1].popularity) - (a[1].popularity > b[1].popularity)

    registry = SortedClassRegistry(
        attr_name="element",
        sort_key=cmp_to_key(compare_pokemon),
    )

    @registry.register
    class Onix(Pokemon):
        element = "rock"
        popularity = 50

    @registry.register
    class Cubone(Pokemon):
        element = "water"
        popularity = 100

    @registry.register
    class Exeggcute(Pokemon):
        element = "grass"
        popularity = 10

    # The registry iterates over registered classes in descending order by
    # ``popularity``.
    assert list(registry.classes()) == [Cubone, Onix, Exeggcute]


def test_gen_lookup_key_overridden():
    """
    When a ``SortedClassRegistry`` overrides the ``gen_lookup_key`` method,
    it can sort by lookup keys if desired.
    """

    def compare_by_lookup_key(a, b):
        """
        :param a: Tuple of (key, class, lookup_key)
        :param b: Tuple of (key, class, lookup_key)
        """
        return (a[2] > b[2]) - (a[2] < b[2])

    class TestRegistry(SortedClassRegistry):
        @staticmethod
        def gen_lookup_key(key: typing.Hashable) -> typing.Hashable:
            """
            Simple override of `gen_lookup_key`, to ensure the sorting
            behaves as expected when the lookup key is different.
            """
            if isinstance(key, str):
                return "".join(reversed(key))
            return key

    registry = TestRegistry(sort_key=cmp_to_key(compare_by_lookup_key))

    registry.register("fire")(Charmander)
    registry.register("grass")(Bulbasaur)
    registry.register("water")(Squirtle)

    assert list(registry.classes()) == [Charmander, Squirtle, Bulbasaur]



"""
Unit tests for :py:class:`class_registry.base.AutoRegister`, which replaces
:py:class:`class_registry.auto_register.AutoRegister` (the latter returns a metaclass,
whilst the former returns a base class).

:see: https://github.com/todofixthis/class-registry/issues/14
"""
from abc import ABC, abstractmethod as abstract_method

import pytest

from class_registry import ClassRegistry
from class_registry.base import AutoRegister


def test_auto_register():
    """
    Using :py:func:`AutoRegister` to, well, auto-register classes.
    """
    registry = ClassRegistry["BasePokemon"](attr_name="element")

    # Note that we declare :py:func:`AutoRegister` as a base class.
    class BasePokemon(AutoRegister(registry), ABC):
        """
        Abstract base class; will not get registered.
        """

        @abstract_method
        def get_abilities(self):
            raise NotImplementedError()

    class Sandslash(BasePokemon):
        """
        Non-abstract subclass; will get registered automatically.
        """

        element = "ground"

        def get_abilities(self):
            return ["sand veil"]

    class BaseEvolvingPokemon(BasePokemon, ABC):
        """
        Abstract subclass; will not get registered.
        """

        @abstract_method
        def evolve(self):
            raise NotImplementedError()

    class Ekans(BaseEvolvingPokemon):
        """
        Non-abstract subclass; will get registered automatically.
        """

        element = "poison"

        def get_abilities(self):
            return ["intimidate", "shed skin"]

        def evolve(self):
            return "Congratulations! Your EKANS evolved into ARBOK!"

    # Note that only non-abstract classes got registered.
    assert list(registry.classes()) == [Sandslash, Ekans]


def test_abstract_strict_definition():
    """
    If a class has no unimplemented abstract methods, it gets registered.
    """
    registry = ClassRegistry["FightingPokemon"](attr_name="element")

    class FightingPokemon(AutoRegister(registry)):
        element = "fighting"

    # :py:class:`FightingPokemon` does not define any abstract methods, so it is not
    # considered to be abstract :shrug:
    assert list(registry.classes()) == [FightingPokemon]


def test_error_attr_name_missing():
    """
    The registry doesn't have an ``attr_name``.
    """
    registry = ClassRegistry()

    with pytest.raises(ValueError):
        AutoRegister(registry)



__all__ = ["RegistryPatcher"]

import typing

from . import RegistryKeyError
from .base import BaseMutableRegistry

T = typing.TypeVar("T")


class RegistryPatcher(typing.Generic[T]):
    """
    Creates a context in which classes are temporarily registered with a class registry,
    then removed when the context exits.

    .. note::

       Only mutable registries can be patched.
    """

    class DoesNotExist(object):
        """
        Used to identify a value that did not exist before we started.
        """

        pass

    def __init__(
        self,
        registry: BaseMutableRegistry[T],
        *args: typing.Type[T],
        **kwargs: typing.Type[T]
    ) -> None:
        """
        :param registry:
            A :py:class:`MutableRegistry` instance to patch.

        :param args:
            Classes to add to the registry.

            This behaves the same as decorating each class with ``@registry.register``.

            .. note::

               ``registry.attr_name`` must be set.

        :param kwargs:
            Same as ``args``, except you explicitly specify the registry keys.

            In the event of a conflict, values in ``args`` override values in
            ``kwargs``.
        """
        super().__init__()

        assert registry.attr_name is not None
        for class_ in args:
            kwargs[getattr(class_, registry.attr_name)] = class_

        self.target: BaseMutableRegistry[T] = registry

        self._new_values: dict[str, typing.Type[T]] = kwargs
        self._prev_values: dict[
            typing.Hashable,
            typing.Union[typing.Type[T], typing.Type[RegistryPatcher.DoesNotExist]],
        ] = {}

    def __enter__(self) -> None:
        self.apply()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.restore()

    def apply(self) -> None:
        """
        Applies the new values.
        """
        # Back up previous values.
        self._prev_values = {
            key: self._get_value(key, self.DoesNotExist) for key in self._new_values
        }

        # Patch values.
        for key, value in self._new_values.items():
            # Remove the existing value first (prevents issues if the registry has
            # ``unique=True``).
            self._del_value(key)

            if value is not self.DoesNotExist:
                self._set_value(key, value)

    def restore(self) -> None:
        """
        Restores previous settings.
        """
        for key, value in self._prev_values.items():
            # Remove the existing value first (prevents issues if the registry has
            # ``unique=True``).
            self._del_value(key)

            if value is not self.DoesNotExist:
                if typing.TYPE_CHECKING:
                    # Convince mypy that ``value`` cannot be ``self.DoesNotExist``.
                    value = typing.cast(typing.Type[T], value)

                self._set_value(key, value)

    def _get_value(self, key: typing.Hashable, default=None) -> typing.Any:
        try:
            return self.target.get_class(key)
        except RegistryKeyError:
            return default

    def _set_value(self, key: typing.Hashable, value: typing.Type[T]) -> None:
        self.target.register(key)(value)

    def _del_value(self, key: typing.Hashable) -> None:
        try:
            self.target.unregister(key)
        except RegistryKeyError:
            pass
